use super::Keyfork;
use clap::{Parser, Subcommand};
use std::path::PathBuf;

use keyfork_mnemonic_util::{English, Mnemonic};
use keyfork_prompt::{default_terminal, DefaultTerminal};
use keyfork_shard::{remote_decrypt, Format};

type Result<T, E = Box<dyn std::error::Error>> = std::result::Result<T, E>;

#[derive(Subcommand, Clone, Debug)]
pub enum RecoverSubcommands {
    /// Decrypt a shard file using keys available on the local system.
    Shard {
        shard_file: PathBuf,
        key_discovery: Option<PathBuf>,
    },

    /// Combine remotely decrypted shards. The shards should be sent using the command `keyfork
    /// shard transport`.
    RemoteShard {},

    /// Read a mnemonic from input.
    Mnemonic {},
}

impl RecoverSubcommands {
    /// Return the 128-bit or 256-bit entropy for a BIP-0039 mnemonic. This is _not_ the same as
    /// the 512-bit seed used by BIP-0032.
    fn handle(&self) -> Result<Vec<u8>> {
        match self {
            RecoverSubcommands::Shard {
                shard_file,
                key_discovery,
            } => {
                let content = std::fs::read_to_string(shard_file)?;
                if content.contains("BEGIN PGP MESSAGE") {
                    let openpgp = keyfork_shard::openpgp::OpenPGP::<DefaultTerminal>::new();
                    let prompt_handler = default_terminal()?;
                    // TODO: remove .clone() by making handle() consume self
                    let seed = openpgp.decrypt_all_shards_to_secret(
                        key_discovery.as_deref(),
                        content.as_bytes(),
                        prompt_handler,
                    )?;
                    Ok(seed)
                } else {
                    panic!("unknown format of shard file");
                }
            }
            RecoverSubcommands::RemoteShard {} => {
                let mut seed = vec![];
                remote_decrypt(&mut seed)?;
                Ok(seed)
            }
            RecoverSubcommands::Mnemonic {} => {
                use keyfork_prompt::{
                    validators::{
                        mnemonic::{MnemonicChoiceValidator, WordLength},
                        Validator,
                    },
                    PromptHandler,
                };
                let mut term = default_terminal()?;
                let validator = MnemonicChoiceValidator {
                    word_lengths: [WordLength::Count(12), WordLength::Count(24)],
                };
                let mnemonic = term.prompt_validated_wordlist::<English, _>(
                    "Mnemonic: ",
                    3,
                    validator.to_fn(),
                )?;
                Ok(mnemonic.to_bytes())
            }
        }
    }
}

#[derive(Parser, Debug, Clone)]
pub struct Recover {
    #[command(subcommand)]
    command: RecoverSubcommands,
}

impl Recover {
    pub fn handle(&self, _k: &Keyfork) -> Result<()> {
        let seed = self.command.handle()?;
        let mnemonic = Mnemonic::from_bytes(&seed)?;
        tokio::runtime::Builder::new_multi_thread()
            .enable_all()
            .build()
            .unwrap()
            .block_on(async {
                keyforkd::setup_registry();
                keyforkd::start_and_run_server(mnemonic).await
            })
    }
}
